@@ -290,10 +290,12 @@ let gen_missing js missing =
 
 let mark_start_of_generated_code = Debug.find ~even_if_quiet:true "mark-runtime-gen"
 
-let link ~standalone ~linkall (js : Javascript.statement_list) : Linker.output =
-  if not standalone
+let link' ~export_runtime ~standalone ~link (js : Javascript.statement_list) :
+    Linker.output =
+  if (not export_runtime) && not standalone
   then { runtime_code = js; always_required_codes = [] }
   else
+    let check_missing = standalone in
     let t = Timer.make () in
     if times () then Format.eprintf "Start Linking...@.";
     let traverse = new Js_traverse.free in
@@ -309,60 +311,121 @@ let link ~standalone ~linkall (js : Javascript.statement_list) : Linker.output =
         :: js
       else js
     in
-    let free = traverse#get_free in
-    let free : StringSet.t =
-      Javascript.IdentSet.fold
-        (fun x acc ->
-          match x with
-          | V _ ->
-              (* This is an error. We don't complain here as we want
-                 to be able to name other variable to make it
-                 easier to spot the problematic ones *)
-              acc
-          | S { name = Utf8 x; _ } -> StringSet.add x acc)
-        free
-        StringSet.empty
+    let used =
+      let all_provided = Linker.list_all () in
+      match link with
+      | `All -> all_provided
+      | `All_from from -> Linker.list_all ~from ()
+      | `No -> StringSet.empty
+      | `Needed ->
+          let free = traverse#get_free in
+          let free : StringSet.t =
+            Javascript.IdentSet.fold
+              (fun x acc ->
+                match x with
+                | V _ ->
+                    (* This is an error. We don't complain here as we want
+                       to be able to name other variable to make it
+                       easier to spot the problematic ones *)
+                    acc
+                | S { name = Utf8 x; _ } -> StringSet.add x acc)
+              free
+              StringSet.empty
+          in
+          let prim = Primitive.get_external () in
+          let all_external = StringSet.union prim all_provided in
+          StringSet.inter free all_external
     in
-    let prim = Primitive.get_external () in
-    let prov = Linker.get_provided () in
-    let all_external = StringSet.union prim prov in
-    let used = StringSet.inter free all_external in
-    let linkinfos = Linker.init () in
-    let linkinfos, missing = Linker.resolve_deps ~linkall linkinfos used in
-    (* gen_missing may use caml_failwith *)
-    let linkinfos, missing =
+    let linkinfos =
+      let from =
+        match link with
+        | `All_from l -> Some l
+        | `All | `No | `Needed -> None
+      in
+      Linker.init ?from ()
+    in
+    let linkinfos, js =
+      let linkinfos, missing = Linker.resolve_deps ~check_missing linkinfos used in
+      (* gen_missing may use caml_failwith *)
       if (not (StringSet.is_empty missing)) && Config.Flag.genprim ()
       then
         let linkinfos, missing2 =
-          Linker.resolve_deps linkinfos (StringSet.singleton "caml_failwith")
+          Linker.resolve_deps
+            ~check_missing
+            linkinfos
+            (StringSet.singleton "caml_failwith")
         in
-        linkinfos, StringSet.union missing missing2
-      else linkinfos, missing
+        let missing = StringSet.union missing missing2 in
+        linkinfos, gen_missing js missing
+      else linkinfos, js
     in
-    let js = if Config.Flag.genprim () then gen_missing js missing else js in
     if times () then Format.eprintf "  linking: %a@." Timer.print t;
     let js =
-      if linkall
+      if export_runtime
       then
         let open Javascript in
-        let all = Linker.all linkinfos in
-        let all =
-          List.map all ~f:(fun name ->
-              let name = Utf8_string.of_string_exn name in
-              Property (PNI name, EVar (ident name)))
-        in
-        ( Expression_statement
-            (EBin
-               ( Eq
-               , dot
-                   (EVar (ident Constant.global_object_))
-                   (Utf8_string.of_string_exn "jsoo_runtime")
-               , EObj all ))
-        , N )
-        :: js
+        match Linker.all linkinfos with
+        | [] -> js
+        | all ->
+            let all =
+              List.map all ~f:(fun name ->
+                  let name = Utf8_string.of_string_exn name in
+                  Property (PNI name, EVar (ident name)))
+            in
+            (if standalone
+             then
+               ( Expression_statement
+                   (EBin
+                      ( Eq
+                      , dot
+                          (EVar (ident Constant.global_object_))
+                          (Utf8_string.of_string_exn "jsoo_runtime")
+                      , EObj all ))
+               , N )
+             else
+               ( Expression_statement
+                   (call
+                      (dot
+                         (EVar (ident (Utf8_string.of_string_exn "Object")))
+                         (Utf8_string.of_string_exn "assign"))
+                      [ dot
+                          (EVar (ident Constant.global_object_))
+                          (Utf8_string.of_string_exn "jsoo_runtime")
+                      ; EObj all
+                      ]
+                      N)
+               , N ))
+            :: js
       else js
     in
-    Linker.link js linkinfos
+    let missing = Linker.missing linkinfos in
+    let output = Linker.link ~check_missing js linkinfos in
+    if not (List.is_empty missing)
+    then
+      { output with
+        runtime_code =
+          (let open Javascript in
+           ( Variable_statement
+               ( Var
+               , [ DeclPattern
+                     ( ObjectBinding
+                         { list =
+                             List.map
+                               ~f:(fun name ->
+                                 let name = Utf8_string.of_string_exn name in
+                                 Prop_ident (Prop_and_ident (ident name), None))
+                               missing
+                         ; rest = None
+                         }
+                     , ( dot
+                           (EVar (ident Constant.global_object_))
+                           (Utf8_string.of_string_exn "jsoo_runtime")
+                       , N ) )
+                 ] )
+           , N )
+           :: output.runtime_code)
+      }
+    else output
 
 let check_js js =
   let t = Timer.make () in
@@ -380,7 +443,7 @@ let check_js js =
       StringSet.empty
   in
   let prim = Primitive.get_external () in
-  let prov = Linker.get_provided () in
+  let prov = Linker.list_all () in
   let all_external = StringSet.union prim prov in
   let missing = StringSet.inter free all_external in
   let missing = StringSet.diff missing Reserved.provided in
@@ -589,8 +652,13 @@ let configure formatter =
   Code.Var.set_pretty (pretty && not (Config.Flag.shortvar ()));
   Code.Var.set_stable (Config.Flag.stable_var ())
 
-let full ~standalone ~wrap_with_fun ~profile ~linkall ~source_map formatter d p =
+let full ~standalone ~wrap_with_fun ~profile ~link ~source_map formatter d p =
   let exported_runtime = not standalone in
+  let export_runtime =
+    match link with
+    | `All | `All_from _ -> true
+    | `Needed | `No -> false
+  in
   let deadcode_sentinal =
     (* If deadcode is disabled, this field is just fresh variable *)
     Code.Var.fresh_n "undef"
@@ -603,7 +671,8 @@ let full ~standalone ~wrap_with_fun ~profile ~linkall ~source_map formatter d p
        | O3 -> o3)
     +> exact_calls ~deadcode_sentinal profile
     +> effects ~deadcode_sentinal
-    +> map_fst (Generate_closure.f +> deadcode')
+    +> map_fst (if Config.Flag.effects () then fun x -> x else Generate_closure.f)
+    +> map_fst deadcode'
   in
   let emit =
     generate
@@ -612,7 +681,7 @@ let full ~standalone ~wrap_with_fun ~profile ~linkall ~source_map formatter d p
       ~wrap_with_fun
       ~warn_on_unhandled_effect:standalone
       ~deadcode_sentinal
-    +> link ~standalone ~linkall
+    +> link' ~export_runtime ~standalone ~link
     +> pack ~wrap_with_fun ~standalone
     +> coloring
     +> check_js
@@ -624,9 +693,9 @@ let full ~standalone ~wrap_with_fun ~profile ~linkall ~source_map formatter d p
   let () = if times () then Format.eprintf " optimizations : %a@." Timer.print t in
   emit r
 
-let full_no_source_map ~standalone ~wrap_with_fun ~profile ~linkall formatter d p =
+let full_no_source_map ~standalone ~wrap_with_fun ~profile ~link formatter d p =
   let (_ : Source_map.t option) =
-    full ~standalone ~wrap_with_fun ~profile ~linkall ~source_map:None formatter d p
+    full ~standalone ~wrap_with_fun ~profile ~link ~source_map:None formatter d p
   in
   ()
 
@@ -634,22 +703,15 @@ let f
     ?(standalone = true)
     ?(wrap_with_fun = `Iife)
     ?(profile = O1)
-    ?(linkall = false)
+    ~link
     ?source_map
     formatter
     d
     p =
-  full ~standalone ~wrap_with_fun ~profile ~linkall ~source_map formatter d p
+  full ~standalone ~wrap_with_fun ~profile ~link ~source_map formatter d p
 
-let f'
-    ?(standalone = true)
-    ?(wrap_with_fun = `Iife)
-    ?(profile = O1)
-    ?(linkall = false)
-    formatter
-    d
-    p =
-  full_no_source_map ~standalone ~wrap_with_fun ~profile ~linkall formatter d p
+let f' ?(standalone = true) ?(wrap_with_fun = `Iife) ?(profile = O1) ~link formatter d p =
+  full_no_source_map ~standalone ~wrap_with_fun ~profile ~link formatter d p
 
 let from_string ~prims ~debug s formatter =
   let p, d = Parse_bytecode.from_string ~prims ~debug s in
@@ -657,7 +719,7 @@ let from_string ~prims ~debug s formatter =
     ~standalone:false
     ~wrap_with_fun:`Anonymous
     ~profile:O1
-    ~linkall:false
+    ~link:`No
     formatter
     d
     p
