@@ -152,11 +152,11 @@ let the_length_of info x =
   get_approx
     info
     (fun x ->
-      match info.info_defs.(Var.idx x) with
-      | Expr (Constant (String s)) -> Some (Int32.of_int (String.length s))
-      | Expr (Prim (Extern "caml_create_string", [ arg ]))
-      | Expr (Prim (Extern "caml_create_bytes", [ arg ])) -> the_int info arg
-      | _ -> None)
+      match Flow.Info.def info x with
+      | Some (Constant (String s)) -> Some (Int32.of_int (String.length s))
+      | Some (Prim (Extern "caml_create_string", [ arg ]))
+      | Some (Prim (Extern "caml_create_bytes", [ arg ])) -> the_int info arg
+      | None | Some _ -> None)
     None
     (fun u v ->
       match u, v with
@@ -175,10 +175,10 @@ let is_int info x =
       get_approx
         info
         (fun x ->
-          match info.info_defs.(Var.idx x) with
-          | Expr (Constant (Int _)) -> Y
-          | Expr (Block (_, _, _)) | Expr (Constant _) -> N
-          | _ -> Unknown)
+          match Flow.Info.def info x with
+          | Some (Constant (Int _)) -> Y
+          | Some (Block (_, _, _, _) | Constant _) -> N
+          | None | Some _ -> Unknown)
         Unknown
         (fun u v ->
           match u, v with
@@ -195,11 +195,15 @@ let the_tag_of info x get =
       get_approx
         info
         (fun x ->
-          match info.info_defs.(Var.idx x) with
-          | Expr (Block (j, _, _)) ->
-              if Var.ISet.mem info.info_possibly_mutable x then None else get j
-          | Expr (Constant (Tuple (j, _, _))) -> get j
-          | _ -> None)
+          match Flow.Info.def info x with
+          | Some (Block (j, _, _, mut)) ->
+              if Flow.Info.possibly_mutable info x
+              then (
+                assert (Poly.(mut = Maybe_mutable));
+                None)
+              else get j
+          | Some (Constant (Tuple (j, _, _))) -> get j
+          | None | Some _ -> None)
         None
         (fun u v ->
           match u, v with
@@ -218,10 +222,10 @@ let the_cont_of info x (a : cont array) =
   get_approx
     info
     (fun x ->
-      match info.info_defs.(Var.idx x) with
-      | Expr (Prim (Extern "%direct_obj_tag", [ b ])) -> the_tag_of info b get
-      | Expr (Constant (Int j)) -> get (Int32.to_int j)
-      | _ -> None)
+      match Flow.Info.def info x with
+      | Some (Prim (Extern "%direct_obj_tag", [ b ])) -> the_tag_of info b get
+      | Some (Constant (Int j)) -> get (Int32.to_int j)
+      | None | Some _ -> None)
     None
     (fun u v ->
       match u, v with
@@ -239,7 +243,7 @@ let eval_instr info ((x, loc) as i) =
           | Some c ->
               let c = if c then 1l else 0l in
               let c = Constant (Int c) in
-              Flow.update_def info x c;
+              Flow.Info.update_def info x c;
               [ Let (x, c), loc ])
       | _ -> [ i ])
   | Let (x, Prim (Extern "caml_ml_string_length", [ s ])) -> (
@@ -253,7 +257,7 @@ let eval_instr info ((x, loc) as i) =
       | None -> [ i ]
       | Some c ->
           let c = Constant (Int c) in
-          Flow.update_def info x c;
+          Flow.Info.update_def info x c;
           [ Let (x, c), loc ])
   | Let (_, Prim (Extern ("caml_array_unsafe_get" | "caml_array_unsafe_set"), _)) ->
       (* Fresh parameters can be introduced for these primitives
@@ -266,19 +270,19 @@ let eval_instr info ((x, loc) as i) =
       | (Y | N) as b ->
           let b = if Poly.(b = N) then 0l else 1l in
           let c = Constant (Int b) in
-          Flow.update_def info x c;
+          Flow.Info.update_def info x c;
           [ Let (x, c), loc ])
   | Let (x, Prim (Extern "%direct_obj_tag", [ y ])) -> (
       match the_tag_of info y (fun x -> Some x) with
       | Some tag ->
           let c = Constant (Int (Int32.of_int tag)) in
-          Flow.update_def info x c;
+          Flow.Info.update_def info x c;
           [ Let (x, c), loc ]
       | None -> [ i ])
   | Let (x, Prim (Extern "caml_sys_const_backend_type", [ _ ])) ->
       let jsoo = Code.Var.fresh () in
       [ Let (jsoo, Constant (String "js_of_ocaml")), noloc
-      ; Let (x, Block (0, [| jsoo |], NotArray)), loc
+      ; Let (x, Block (0, [| jsoo |], NotArray, Immutable)), loc
       ]
   | Let (_, Prim (Extern ("%resume" | "%perform" | "%reperform"), _)) ->
       [ i ] (* We need that the arguments to this primitives remain variables *)
@@ -299,7 +303,7 @@ let eval_instr info ((x, loc) as i) =
       match res with
       | Some c ->
           let c = Constant c in
-          Flow.update_def info x c;
+          Flow.Info.update_def info x c;
           [ Let (x, c), loc ]
       | _ ->
           [ ( Let
@@ -327,9 +331,9 @@ let the_cond_of info x =
   get_approx
     info
     (fun x ->
-      match info.info_defs.(Var.idx x) with
-      | Expr (Constant (Int 0l)) -> Zero
-      | Expr
+      match Flow.Info.def info x with
+      | Some (Constant (Int 0l)) -> Zero
+      | Some
           (Constant
             ( Int _
             | Float _
@@ -338,9 +342,9 @@ let the_cond_of info x =
             | NativeString _
             | Float_array _
             | Int64 _ )) -> Non_zero
-      | Expr (Block (_, _, _)) -> Non_zero
-      | Expr (Field _ | Closure _ | Prim _ | Apply _) -> Unknown
-      | Param | Phi _ -> Unknown)
+      | Some (Block (_, _, _, _)) -> Non_zero
+      | Some (Field _ | Closure _ | Prim _ | Apply _ | Special _) -> Unknown
+      | None -> Unknown)
     Unknown
     (fun u v ->
       match u, v with
@@ -381,8 +385,9 @@ let rec do_not_raise pc visited blocks =
         | Array_set (_, _, _) | Offset_ref (_, _) | Set_field (_, _, _) | Assign _ -> ()
         | Let (_, e) -> (
             match e with
-            | Block (_, _, _) | Field (_, _) | Constant _ | Closure _ -> ()
+            | Block (_, _, _, _) | Field (_, _) | Constant _ | Closure _ -> ()
             | Apply _ -> raise May_raise
+            | Special _ -> ()
             | Prim (Extern name, _) when Primitive.is_pure name -> ()
             | Prim (Extern _, _) -> raise May_raise
             | Prim (_, _) -> ()));
@@ -406,8 +411,7 @@ let drop_exception_handler blocks =
   Addr.Map.fold
     (fun pc _ blocks ->
       match Addr.Map.find pc blocks with
-      | { branch = Pushtrap (((addr, _) as cont1), _x, _cont2, addrset), loc; _ } as b
-        -> (
+      | { branch = Pushtrap (((addr, _) as cont1), _x, _cont2), loc; _ } as b -> (
           try
             let visited = do_not_raise addr Addr.Set.empty blocks in
             let b = { b with branch = Branch cont1, loc } in
@@ -418,9 +422,7 @@ let drop_exception_handler blocks =
                   let b = Addr.Map.find pc2 blocks in
                   let branch =
                     match b.branch with
-                    | Poptrap ((addr, _) as cont), loc ->
-                        assert (Addr.Set.mem addr addrset);
-                        Branch cont, loc
+                    | Poptrap cont, loc -> Branch cont, loc
                     | x -> x
                   in
                   let b = { b with branch } in
