@@ -18,16 +18,6 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *)
 
-(*XXX
-  Patterns:
-  => should have special code for switches that include the preceding
-     if statement when possible
-  => if e1 then {if e2 then P else Q} else {if e3 then P else Q}
-  => if e then return e1; return e2
-  => if e then var x = e1; else var x = e2;
-  => while (true) {.... if (e) continue; break; }
-*)
-
 open! Stdlib
 
 let debug = Debug.find "gen"
@@ -42,19 +32,19 @@ module J = Javascript
 let string_of_set s =
   String.concat ~sep:", " (List.map ~f:Addr.to_string (Addr.Set.elements s))
 
-let rec list_group_rec f g l b m n =
+let rec list_group_rec ~equal f g l b m n =
   match l with
   | [] -> List.rev ((b, List.rev m) :: n)
   | a :: r ->
       let fa = f a in
-      if Poly.(fa = b)
-      then list_group_rec f g r b (g a :: m) n
-      else list_group_rec f g r fa [ g a ] ((b, List.rev m) :: n)
+      if equal fa b
+      then list_group_rec ~equal f g r b (g a :: m) n
+      else list_group_rec ~equal f g r fa [ g a ] ((b, List.rev m) :: n)
 
-let list_group f g l =
+let list_group ~equal f g l =
   match l with
   | [] -> []
-  | a :: r -> list_group_rec f g r (f a) [ g a ] []
+  | a :: r -> list_group_rec ~equal f g r (f a) [ g a ] []
 
 (****)
 
@@ -164,7 +154,7 @@ module Share = struct
                   if (not exact) || cps
                   then add_apply { arity = List.length args; exact; cps } share
                   else share
-              | Let (_, Prim (Extern "%closure", [ Pc (String name) ])) ->
+              | Let (_, Special (Alias_prim name)) ->
                   let name = Primitive.resolve name in
                   let share =
                     if Primitive.exists name then add_prim name share else share
@@ -281,6 +271,8 @@ module Ctx = struct
     ; effect_warning : bool ref
     ; cps_calls : Effects.cps_calls
     ; deadcode_sentinal : Var.t
+    ; mutated_vars : Code.Var.Set.t Code.Addr.Map.t
+    ; freevars : Code.Var.Set.t Code.Addr.Map.t
     }
 
   let initial
@@ -288,6 +280,8 @@ module Ctx = struct
       ~exported_runtime
       ~should_export
       ~deadcode_sentinal
+      ~mutated_vars
+      ~freevars
       blocks
       live
       cps_calls
@@ -302,6 +296,8 @@ module Ctx = struct
     ; effect_warning = ref (not warn_on_unhandled_effect)
     ; cps_calls
     ; deadcode_sentinal
+    ; mutated_vars
+    ; freevars
     }
 end
 
@@ -501,7 +497,6 @@ let constant ~ctx x level =
 
 type queue_elt =
   { prop : int
-  ; cardinal : int
   ; ce : J.expression
   ; loc : J.location
   ; deps : Code.Var.Set.t
@@ -510,15 +505,15 @@ type queue_elt =
 let access_queue queue x =
   try
     let elt = List.assoc x queue in
-    if elt.cardinal = 1
-    then ((elt.prop, elt.deps), elt.ce), List.remove_assoc x queue
-    else
-      ( ((elt.prop, elt.deps), elt.ce)
-      , List.map queue ~f:(function
-            | x', elt when Var.equal x x' -> x', { elt with cardinal = pred elt.cardinal }
-            | x -> x) )
+    ((elt.prop, elt.deps), elt.ce), List.remove_assoc x queue
   with Not_found -> ((fst const_p, Code.Var.Set.singleton x), var x), queue
 
+let access_queue_loc queue x =
+  try
+    let elt = List.assoc x queue in
+    ((elt.prop, elt.deps), elt.ce, elt.loc), List.remove_assoc x queue
+  with Not_found -> ((fst const_p, Code.Var.Set.singleton x), var x, J.N), queue
+
 let access_queue' ~ctx queue x =
   match x with
   | Pc c ->
@@ -528,22 +523,6 @@ let access_queue' ~ctx queue x =
       (const_p, js), queue
   | Pv x -> access_queue queue x
 
-let access_queue_may_flush queue v x =
-  let tx, queue = access_queue queue x in
-  let _, instrs, queue =
-    List.fold_left
-      queue
-      ~init:(Code.Var.Set.singleton v, [], [])
-      ~f:(fun (deps, instrs, queue) ((y, elt) as eq) ->
-        if not (Code.Var.Set.disjoint deps elt.deps)
-        then
-          ( Code.Var.Set.add y deps
-          , (J.variable_declaration [ J.V y, (elt.ce, elt.loc) ], elt.loc) :: instrs
-          , queue )
-        else deps, instrs, eq :: queue)
-  in
-  instrs, (tx, List.rev queue)
-
 let should_flush (cond, _) prop = cond <> fst const_p && cond + prop >= fst flush_p
 
 let flush_queue expr_queue prop (l : J.statement_list) =
@@ -560,18 +539,14 @@ let flush_queue expr_queue prop (l : J.statement_list) =
 
 let flush_all expr_queue l = fst (flush_queue expr_queue flush_p l)
 
-let enqueue expr_queue prop x ce loc cardinal acc =
+let enqueue expr_queue prop x ce loc acc =
   let instrs, expr_queue =
     if Config.Flag.compact ()
-    then if is_mutable prop then flush_queue expr_queue prop [] else [], expr_queue
-    else flush_queue expr_queue flush_p []
+    then if is_mutable prop then flush_queue expr_queue prop acc else acc, expr_queue
+    else flush_queue expr_queue flush_p acc
   in
   let prop, deps = prop in
-  let deps =
-    List.fold_left expr_queue ~init:deps ~f:(fun deps (x', elt) ->
-        if Code.Var.Set.mem x' deps then Code.Var.Set.union elt.deps deps else deps)
-  in
-  instrs @ acc, (x, { prop; deps; ce; loc; cardinal }) :: expr_queue
+  instrs, (x, { prop; deps; ce; loc }) :: expr_queue
 
 (****)
 
@@ -580,11 +555,10 @@ type state =
   ; dom : Structure.graph
   ; visited_blocks : Addr.Set.t ref
   ; ctx : Ctx.t
-  ; blocks : Code.block Addr.Map.t
   }
 
 module DTree = struct
-  (* This as to be kept in sync with the way we build conditionals
+  (* This has to be kept in sync with the way we build conditionals
      and switches! *)
 
   type cond =
@@ -604,7 +578,7 @@ module DTree = struct
     a
     |> Array.to_list
     |> List.sort ~cmp:(fun (cont1, _) (cont2, _) -> Poly.compare cont1 cont2)
-    |> list_group fst snd
+    |> list_group ~equal:Poly.equal fst snd
     |> List.map ~f:(fun (cont1, l1) -> cont1, List.flatten l1)
     |> List.sort ~cmp:(fun (_, l1) (_, l2) -> compare (List.length l1) (List.length l2))
     |> Array.of_list
@@ -616,7 +590,7 @@ module DTree = struct
     let ai = Array.mapi a ~f:(fun i x -> x, i) in
     (* group the contiguous cases with the same continuation *)
     let ai : (Code.cont * int list) array =
-      Array.of_list (list_group fst snd (Array.to_list ai))
+      Array.of_list (list_group ~equal:Poly.equal fst snd (Array.to_list ai))
     in
     let rec loop low up =
       let array_norm : (Code.cont * int list) array =
@@ -699,10 +673,9 @@ end
 
 let build_graph ctx pc =
   let visited_blocks = ref Addr.Set.empty in
-  let blocks = ctx.Ctx.blocks in
-  let structure = Structure.build_graph blocks pc in
+  let structure = Structure.build_graph ctx.Ctx.blocks pc in
   let dom = Structure.dominator_tree structure in
-  { visited_blocks; structure; dom; ctx; blocks }
+  { visited_blocks; structure; dom; ctx }
 
 (****)
 
@@ -739,22 +712,64 @@ let visit_all params args =
   in
   l
 
-let parallel_renaming params args continuation queue =
-  let l = List.rev (visit_all params args) in
-  List.fold_left
-    l
-    ~f:(fun continuation (y, x) queue ->
-      let instrs, ((px, cx), queue) = access_queue_may_flush queue y x in
-      let st, queue =
-        flush_queue
-          queue
-          px
-          (instrs @ [ J.variable_declaration [ J.V y, (cx, J.N) ], J.N ])
-      in
-      let never, code = continuation queue in
-      never, st @ code)
-    ~init:continuation
-    queue
+let parallel_renaming back_edge params args continuation queue =
+  if back_edge && Config.Flag.es6 ()
+     (* This is likely slower than using explicit temp variable
+        but let's experiment with es6 a bit *)
+  then
+    let args, params =
+      List.map2 args params ~f:(fun a p -> if Var.equal a p then None else Some (a, p))
+      |> List.filter_map ~f:(fun x -> x)
+      |> List.split
+    in
+    let args, _, queue =
+      List.fold_left args ~init:([], const_p, queue) ~f:(fun (acc, p, queue) a ->
+          let (px, cx), queue = access_queue queue a in
+          cx :: acc, or_p px p, queue)
+    in
+    let never, code = continuation queue in
+    match params, args with
+    | [ p ], [ a ] ->
+        never, (J.Expression_statement (J.EBin (J.Eq, J.EVar (J.V p), a)), J.N) :: code
+    | params, args ->
+        let lhs =
+          J.EAssignTarget
+            (J.ArrayTarget (List.map params ~f:(fun p -> J.TargetElementId (J.V p, None))))
+        in
+        let rhs = J.EArr (List.rev_map args ~f:(fun x -> J.Element x)) in
+        never, (J.Expression_statement (J.EBin (J.Eq, lhs, rhs)), J.N) :: code
+  else
+    let l = visit_all params args in
+    (* if not back_edge
+     * then assert (Poly.( = ) l (List.rev_map2 params args ~f:(fun a b -> a, b))); *)
+    let queue, before, renaming, _ =
+      List.fold_left
+        l
+        ~init:(queue, [], [], Code.Var.Set.empty)
+        ~f:(fun (queue, before, renaming, seen) (y, x) ->
+          let ((_, deps_x), cx, locx), queue = access_queue_loc queue x in
+          let seen' = Code.Var.Set.add y seen in
+          if not Code.Var.Set.(is_empty (inter seen deps_x))
+          then
+            let () = assert back_edge in
+            let before = (J.variable_declaration [ J.V x, (cx, locx) ], locx) :: before in
+            let renaming = (y, J.EVar (J.V x)) :: renaming in
+            queue, before, renaming, seen'
+          else
+            let renaming = (y, cx) :: renaming in
+            queue, before, renaming, seen')
+    in
+    let renaming =
+      if back_edge
+      then
+        List.map renaming ~f:(fun (t, e) ->
+            J.Expression_statement (J.EBin (J.Eq, J.EVar (J.V t), e)), J.N)
+      else
+        List.map renaming ~f:(fun (t, e) ->
+            J.variable_declaration [ J.V t, (e, J.N) ], J.N)
+    in
+    let never, code = continuation queue in
+    never, List.rev_append before (List.rev_append renaming code)
 
 (****)
 
@@ -912,8 +927,16 @@ let _ =
   register_un_prim "%direct_obj_tag" `Mutator (fun cx _loc -> Mlvalue.Block.tag cx);
   register_bin_prim "caml_array_unsafe_get" `Mutable (fun cx cy _ ->
       Mlvalue.Array.field cx cy);
-  register_bin_prim "%int_add" `Pure (fun cx cy _ -> to_int (plus_int cx cy));
-  register_bin_prim "%int_sub" `Pure (fun cx cy _ -> to_int (J.EBin (J.Minus, cx, cy)));
+  register_bin_prim "%int_add" `Pure (fun cx cy _ ->
+      match cx, cy with
+      | J.EBin (J.Minus, cz, J.ENum n), J.ENum m ->
+          to_int (J.EBin (J.Plus, cz, J.ENum (J.Num.add m (J.Num.neg n))))
+      | _ -> to_int (plus_int cx cy));
+  register_bin_prim "%int_sub" `Pure (fun cx cy _ ->
+      match cx, cy with
+      | J.EBin (J.Minus, cz, J.ENum n), J.ENum m ->
+          to_int (J.EBin (J.Minus, cz, J.ENum (J.Num.add n m)))
+      | _ -> to_int (J.EBin (J.Minus, cx, cy)));
   register_bin_prim "%direct_int_mul" `Pure (fun cx cy _ ->
       to_int (J.EBin (J.Mul, cx, cy)));
   register_bin_prim "%direct_int_div" `Pure (fun cx cy _ ->
@@ -979,17 +1002,6 @@ let _ =
       bool (J.EBin (J.InstanceOf, cx, cy)));
   register_un_prim "caml_js_typeof" `Mutator (fun cx _ -> J.EUn (J.Typeof, cx))
 
-(* This is not correct when switching the js-string flag *)
-(* {[
-    register_un_prim "caml_jsstring_of_string" `Mutable (fun cx loc ->
-      J.ECall (J.EDot (cx, "toString"), [], loc));
-    register_bin_prim "caml_string_notequal" `Pure (fun cx cy _ ->
-      J.EBin (J.NotEqEq, cx, cy));
-    register_bin_prim "caml_string_equal" `Pure (fun cx cy _ ->
-      bool (J.EBin (J.EqEq, cx, cy)))
-     ]}
-*)
-
 (****)
 (* when raising ocaml exception and [improved_stacktrace] is enabled,
    tag the ocaml exception with a Javascript error (that contain js stacktrace).
@@ -1030,7 +1042,7 @@ let rec translate_expr ctx queue loc x e level : _ * J.statement_list =
       let prop = or_p prop prop' in
       let e = apply_fun ctx f args exact cps loc in
       (e, prop, queue), []
-  | Block (tag, a, array_or_not) ->
+  | Block (tag, a, array_or_not, _mut) ->
       let contents, prop, queue =
         List.fold_right
           ~f:(fun x (args, prop, queue) ->
@@ -1058,6 +1070,7 @@ let rec translate_expr ctx queue loc x e level : _ * J.statement_list =
       (Mlvalue.Block.field cx n, or_p px mutable_p, queue), []
   | Closure (args, ((pc, _) as cont)) ->
       let loc = source_location ctx ~force:After (After pc) in
+      let fv = Addr.Map.find pc ctx.freevars in
       let clo = compile_closure ctx cont in
       let clo =
         match clo with
@@ -1073,10 +1086,15 @@ let rec translate_expr ctx queue loc x e level : _ * J.statement_list =
         | _ -> clo
       in
       let clo = J.EFun (None, J.fun_ (List.map args ~f:(fun v -> J.V v)) clo loc) in
-      (clo, flush_p, queue), []
+      (clo, (fst const_p, fv), queue), []
   | Constant c ->
       let js, instrs = constant ~ctx c level in
       (js, const_p, queue), instrs
+  | Special (Alias_prim name) ->
+      let prim = Share.get_prim (runtime_fun ctx) name ctx.Ctx.share in
+      (prim, const_p, queue), []
+  | Special Undefined ->
+      (J.(EVar (ident (Utf8_string.of_string_exn "undefined"))), const_p, queue), []
   | Prim (Extern "debugger", _) ->
       let ins =
         if Config.Flag.debugger () then J.Debugger_statement else J.Empty_statement
@@ -1135,13 +1153,6 @@ let rec translate_expr ctx queue loc x e level : _ * J.statement_list =
                 ~init:([], const_p, queue)
             in
             J.array args, prop, queue
-        | Extern "%closure", [ Pc (String name) ] ->
-            let prim = Share.get_prim (runtime_fun ctx) name ctx.Ctx.share in
-            prim, const_p, queue
-        | Extern "%closure", _ -> assert false
-        | Extern "%undefined", [] ->
-            J.(EVar (ident (Utf8_string.of_string_exn "undefined"))), const_p, queue
-        | Extern "%undefined", _ -> assert false
         | Extern "%caml_js_opt_call", f :: o :: l ->
             let (pf, cf), queue = access_queue' ~ctx queue f in
             let (po, co), queue = access_queue' ~ctx queue o in
@@ -1210,11 +1221,8 @@ let rec translate_expr ctx queue loc x e level : _ * J.statement_list =
            | Extern "caml_js_get", [ _; Pc (String _) ] -> assert false
            | Extern "caml_js_set", [ _; Pc (String s); _ ] -> assert false
            | Extern "caml_js_delete", [ _; Pc (String _) ] -> assert false
-         ]}
-      *)
-        | Extern "%overrideMod", [ Pc (String m); Pc (String f) ] ->
-            runtime_fun ctx (Printf.sprintf "caml_%s_%s" m f), const_p, queue
-        | Extern "%overrideMod", _ -> assert false
+           ]}
+        *)
         | Extern "%caml_js_opt_object", fields ->
             let rec build_fields queue l =
               match l with
@@ -1260,6 +1268,26 @@ let rec translate_expr ctx queue loc x e level : _ * J.statement_list =
             let prim = Share.get_prim (runtime_fun ctx) name ctx.Ctx.share in
             let prim_kind = kind (Primitive.kind name) in
             J.call prim [] loc, prim_kind, queue
+        | Extern "caml_string_notequal", [ a; b ] when Config.Flag.use_js_string () ->
+            let (px, cx), queue = access_queue' ~ctx queue a in
+            let (py, cy), queue = access_queue' ~ctx queue b in
+            let prop = or_p px py in
+            bool (J.EBin (J.NotEqEq, cx, cy)), prop, queue
+        | Extern "caml_string_equal", [ a; b ] when Config.Flag.use_js_string () ->
+            let (px, cx), queue = access_queue' ~ctx queue a in
+            let (py, cy), queue = access_queue' ~ctx queue b in
+            let prop = or_p px py in
+            bool (J.EBin (J.EqEq, cx, cy)), prop, queue
+        | Extern "caml_string_concat", [ a; b ] when Config.Flag.use_js_string () ->
+            let (pa, ca), queue = access_queue' ~ctx queue a in
+            let (pb, cb), queue = access_queue' ~ctx queue b in
+            let prop = or_p pa pb in
+            let rec add ca cb =
+              match cb with
+              | J.EBin (J.Plus, cb1, cb2) -> J.EBin (J.Plus, add ca cb1, cb2)
+              | _ -> J.EBin (J.Plus, ca, cb)
+            in
+            add ca cb, prop, queue
         | Extern name, l -> (
             let name = Primitive.resolve name in
             match internal_prim name with
@@ -1325,11 +1353,12 @@ and translate_instr ctx expr_queue instr =
       let keep_name x =
         match Code.Var.get_name x with
         | None -> false
-        (* "switcher" is emitted by the OCaml compiler when compiling
-           pattern matching, it does not help much to keep it in the
-           generated js, let's drop it *)
         | Some "" -> false
-        | Some s -> (not (generated_name s)) && not (String.is_prefix s ~prefix:"jsoo_")
+        | Some s ->
+            (* "switcher" is emitted by the OCaml compiler when compiling
+               pattern matching, it does not help much to keep it in the
+               generated js, let's drop it *)
+            (not (generated_name s)) && not (String.is_prefix s ~prefix:"jsoo_")
       in
       match ctx.Ctx.live.(Var.idx x), e with
       | 0, _ ->
@@ -1337,14 +1366,8 @@ and translate_instr ctx expr_queue instr =
           flush_queue expr_queue prop (instrs @ [ J.Expression_statement ce, loc ])
       | 1, _
         when Config.Flag.compact () && ((not (Config.Flag.pretty ())) || not (keep_name x))
-        -> enqueue expr_queue prop x ce loc 1 instrs
-      (* We could inline more.
-         size_v : length of the variable after serialization
-         size_c : length of the constant after serialization
-         num : number of occurrence
-         size_c * n < size_v * n + size_v + 1 + size_c
-      *)
-      | n, Constant (Int _ | Float _) -> enqueue expr_queue prop x ce loc n instrs
+        -> enqueue expr_queue prop x ce loc instrs
+      | 1, Constant (Int _ | Float _) -> enqueue expr_queue prop x ce loc instrs
       | _ ->
           flush_queue
             expr_queue
@@ -1358,23 +1381,19 @@ and translate_instr ctx expr_queue instr =
         expr_queue
         mutator_p
         [ J.Expression_statement (J.EBin (J.Eq, Mlvalue.Block.field cx n, cy)), loc ]
-  | Offset_ref (x, 1) ->
-      let loc = source_location ctx pc in
-      (* FIX: may overflow.. *)
-      let (_px, cx), expr_queue = access_queue expr_queue x in
-      flush_queue
-        expr_queue
-        mutator_p
-        [ J.Expression_statement (J.EUn (J.IncrA, Mlvalue.Block.field cx 0)), loc ]
   | Offset_ref (x, n) ->
       let loc = source_location ctx pc in
       (* FIX: may overflow.. *)
       let (_px, cx), expr_queue = access_queue expr_queue x in
-      flush_queue
-        expr_queue
-        mutator_p
-        [ J.Expression_statement (J.EBin (J.PlusEq, Mlvalue.Block.field cx 0, int n)), loc
-        ]
+      let expr = Mlvalue.Block.field cx 0 in
+      let expr' =
+        match n with
+        | 1 -> J.EUn (J.IncrA, expr)
+        | -1 -> J.EUn (J.DecrA, expr)
+        | n when n < 0 (* *) -> J.EBin (J.MinusEq, expr, int (-n))
+        | n (*   n > 0    *) -> J.EBin (J.PlusEq, expr, int n)
+      in
+      flush_queue expr_queue mutator_p [ J.Expression_statement expr', loc ]
   | Array_set (x, y, z) ->
       let loc = source_location ctx pc in
       let (_px, cx), expr_queue = access_queue expr_queue x in
@@ -1385,18 +1404,134 @@ and translate_instr ctx expr_queue instr =
         mutator_p
         [ J.Expression_statement (J.EBin (J.Eq, Mlvalue.Array.field cx cy, cz)), loc ]
 
-and translate_instrs ctx expr_queue instr last =
-  match instr with
-  | [] -> [], expr_queue
+and translate_instrs_rev (ctx : Ctx.t) expr_queue instrs acc_rev muts_map : _ * _ =
+  match instrs with
+  | [] -> acc_rev, expr_queue
+  | (Let (_, Closure _), _) :: _ ->
+      let names, pcs, all, rem = collect_closures instrs in
+      let fvs =
+        List.fold_left pcs ~init:Code.Var.Set.empty ~f:(fun acc pc ->
+            Code.Var.Set.union acc (Addr.Map.find pc ctx.freevars))
+      in
+      let muts =
+        List.fold_left pcs ~init:Code.Var.Set.empty ~f:(fun acc pc ->
+            Code.Var.Set.union acc (Code.Addr.Map.find pc ctx.Ctx.mutated_vars))
+      in
+      let names =
+        List.fold_left names ~init:Code.Var.Set.empty ~f:(fun acc name ->
+            Code.Var.Set.add name acc)
+      in
+      assert (Code.Var.Set.cardinal names = List.length all);
+      assert (Code.Var.Set.(is_empty (diff muts fvs)));
+      let old_muts_map = muts_map in
+      let muts_map_l =
+        Code.Var.Set.elements muts
+        |> List.map ~f:(fun x ->
+               ( x
+               , match Code.Var.Map.find_opt x old_muts_map with
+                 | None -> Code.Var.fork x
+                 | Some x' -> x' ))
+      in
+      let muts_map =
+        List.fold_left muts_map_l ~init:old_muts_map ~f:(fun acc (x, x') ->
+            Var.Map.add x x' acc)
+      in
+      (* Rewrite blocks using well-scoped closure variables *)
+      let ctx =
+        if List.is_empty muts_map_l
+        then ctx
+        else
+          let subst = Subst.from_map muts_map in
+          let p, _visited =
+            List.fold_left
+              pcs
+              ~init:(ctx.blocks, Addr.Set.empty)
+              ~f:(fun (blocks, visited) pc -> Subst.cont' subst pc blocks visited)
+          in
+          { ctx with blocks = p }
+      in
+      let vd kind = function
+        | [] -> []
+        | l -> [ J.variable_declaration ~kind (List.rev l), J.N ]
+      in
+      (* flush variables part of closures env from the queue *)
+      let bind_fvs, bind_fvs_muts, expr_queue =
+        let expr_queue, vars, lets =
+          Code.Var.Set.fold
+            (fun v (expr_queue, vars, lets) ->
+              assert (not (Code.Var.Set.mem v names));
+              let (px, cx, locx), expr_queue = access_queue_loc expr_queue v in
+              let flushed = Code.Var.Set.(equal (snd px) (singleton v)) in
+              match
+                ( flushed
+                , Code.Var.Map.find_opt v muts_map
+                , Code.Var.Map.find_opt v old_muts_map )
+              with
+              | true, None, _ -> expr_queue, vars, lets
+              | (true | false), Some _, Some _ -> expr_queue, vars, lets
+              | (true | false), Some v', None ->
+                  let lets = (J.V v', (cx, locx)) :: lets in
+                  expr_queue, vars, lets
+              | false, None, _ ->
+                  let vars = (J.V v, (cx, locx)) :: vars in
+                  expr_queue, vars, lets)
+            (Code.Var.Set.diff fvs names)
+            (expr_queue, [], [])
+        in
+        vars, lets, expr_queue
+      in
+      (* Mutually recursive functions need to be properly scoped. *)
+      let bind_fvs_rec, funs_rev, expr_queue =
+        List.fold_left
+          all
+          ~init:([], [], expr_queue)
+          ~f:(fun (mut_rec, st_rev, expr_queue) i ->
+            let x' =
+              match i with
+              | Let (x', _), _ -> x'
+              | _ -> assert false
+            in
+            let l, expr_queue = translate_instr ctx expr_queue i in
+            if Code.Var.Set.mem x' fvs
+            then
+              let mut_rec =
+                match Code.Var.Map.find_opt x' muts_map with
+                | None -> mut_rec
+                | Some v' -> (J.V v', (J.EVar (J.V x'), J.N)) :: mut_rec
+              in
+              match l with
+              | [ i ] -> mut_rec, i :: st_rev, expr_queue
+              | [] ->
+                  let (_px, cx, locx), expr_queue = access_queue_loc expr_queue x' in
+                  ( mut_rec
+                  , (J.variable_declaration [ J.V x', (cx, locx) ], locx) :: st_rev
+                  , expr_queue )
+              | _ :: _ :: _ -> assert false
+            else mut_rec, List.rev_append l st_rev, expr_queue)
+      in
+      let acc_rev = vd Var bind_fvs @ acc_rev in
+      let acc_rev = vd Let bind_fvs_muts @ acc_rev in
+      let acc_rev = funs_rev @ acc_rev in
+      let acc_rev = vd Let bind_fvs_rec @ acc_rev in
+      translate_instrs_rev ctx expr_queue rem acc_rev muts_map
   | instr :: rem ->
       let st, expr_queue = translate_instr ctx expr_queue instr in
-      let instrs, expr_queue = translate_instrs ctx expr_queue rem last in
-      st @ instrs, expr_queue
+      let acc_rev = List.rev_append st acc_rev in
+      translate_instrs_rev ctx expr_queue rem acc_rev muts_map
+
+and translate_instrs (ctx : Ctx.t) expr_queue instrs =
+  let st_rev, expr_queue =
+    translate_instrs_rev (ctx : Ctx.t) expr_queue instrs [] Var.Map.empty
+  in
+  List.rev st_rev, expr_queue
 
 (* Compile loops. *)
 and compile_block st queue (pc : Addr.t) scope_stack ~fall_through =
   if (not (List.is_empty queue))
-     && (Structure.is_loop_header st.structure pc || not (Config.Flag.inline ()))
+     && (Structure.is_loop_header st.structure pc
+        || (* Do not inline expressions across block boundaries when --no-inline is used
+              Single-stepping in the debugger should work better this way (fixes #290). *)
+        not (Config.Flag.inline ()))
   then
     let never, code = compile_block st [] pc scope_stack ~fall_through in
     never, flush_all queue code
@@ -1443,8 +1578,8 @@ and compile_block_no_loop st queue (pc : Addr.t) ~fall_through scope_stack =
     assert false);
   if debug () then Format.eprintf "Compiling block %d@;" pc;
   st.visited_blocks := Addr.Set.add pc !(st.visited_blocks);
-  let block = Addr.Map.find pc st.blocks in
-  let seq, queue = translate_instrs st.ctx queue block.body block.branch in
+  let block = Addr.Map.find pc st.ctx.blocks in
+  let seq, queue = translate_instrs st.ctx queue block.body in
   let nbbranch =
     match fst block.branch with
     | Switch (_, a) ->
@@ -1457,7 +1592,6 @@ and compile_block_no_loop st queue (pc : Addr.t) ~fall_through scope_stack =
         fun pc -> DTree.nbbranch dtree pc
     | _ -> fun _pc -> 0
   in
-
   let new_scopes =
     Structure.get_edges st.dom pc
     |> Addr.Set.elements
@@ -1604,7 +1738,7 @@ and compile_conditional st queue ~fall_through last scope_stack : _ * _ =
         in
         true, flush_all queue [ J.Return_statement e_opt, loc ]
     | Branch cont -> compile_branch st queue cont scope_stack ~fall_through
-    | Pushtrap (c1, x, e1, _) ->
+    | Pushtrap (c1, x, e1) ->
         let never_body, body = compile_branch st [] c1 scope_stack ~fall_through in
         if debug () then Format.eprintf "@,}@]@,@[<hv 2>catch {@;";
         let never_handler, handler = compile_branch st [] e1 scope_stack ~fall_through in
@@ -1668,21 +1802,29 @@ and compile_conditional st queue ~fall_through last scope_stack : _ * _ =
      | Switch _ | Cond _ | Pushtrap _ -> Format.eprintf "}@]@;");
   res
 
-and compile_argument_passing ctx queue (pc, args) continuation =
+and compile_argument_passing ctx queue (pc, args) back_edge continuation =
   if List.is_empty args
   then continuation queue
   else
     let block = Addr.Map.find pc ctx.Ctx.blocks in
-    parallel_renaming block.params args continuation queue
+    parallel_renaming back_edge block.params args continuation queue
 
 and compile_branch st queue ((pc, _) as cont) scope_stack ~fall_through : bool * _ =
-  compile_argument_passing st.ctx queue cont (fun queue ->
+  let scope = List.assoc_opt pc scope_stack in
+  let back_edge =
+    List.exists
+      ~f:(function
+        | pc', (_, _, Loop) when pc' = pc -> true
+        | _ -> false)
+      scope_stack
+  in
+  compile_argument_passing st.ctx queue cont back_edge (fun queue ->
       if match fall_through with
          | Block pc' -> pc' = pc
          | Return -> false
       then false, flush_all queue []
       else
-        match List.assoc_opt pc scope_stack with
+        match scope with
         | Some (l, used, Loop) ->
             (* Loop back to the beginning of the loop using continue.
                We can skip the label if we're not inside a nested loop. *)
@@ -1752,6 +1894,13 @@ and compile_closure ctx (pc, args) =
   if debug () then Format.eprintf "}@]@;";
   res
 
+and collect_closures l =
+  match l with
+  | ((Let (x, Closure (_, (pc, _))), _loc) as i) :: rem ->
+      let names', pcs', i', rem' = collect_closures rem in
+      x :: names', pc :: pcs', i :: i', rem'
+  | _ -> [], [], [], l
+
 let generate_shared_value ctx =
   let strings =
     ( J.variable_declaration
@@ -1810,12 +1959,16 @@ let f
   let exported_runtime =
     if exported_runtime then Some (Code.Var.fresh_n "runtime", ref false) else None
   in
+  let mutated_vars = Freevars.f_mutable p in
+  let freevars = Freevars.f p in
   let ctx =
     Ctx.initial
       ~warn_on_unhandled_effect
       ~exported_runtime
       ~should_export
       ~deadcode_sentinal
+      ~mutated_vars
+      ~freevars
       p.blocks
       live_vars
       cps_calls
