@@ -367,7 +367,7 @@ and block = statement_list
 and statement_list = (statement * location) list
 
 and variable_declaration =
-  | DeclIdent of binding_ident * initialiser option
+  | DeclIdent of ident * initialiser option
   | DeclPattern of binding_pattern * initialiser
 
 and variable_declaration_kind =
@@ -416,16 +416,16 @@ and for_binding = binding
 and binding_element = binding * initialiser option
 
 and binding =
-  | BindingIdent of binding_ident
+  | BindingIdent of ident
   | BindingPattern of binding_pattern
 
 and binding_pattern =
-  | ObjectBinding of (binding_property, binding_ident) list_with_rest
+  | ObjectBinding of (binding_property, ident) list_with_rest
   | ArrayBinding of (binding_element option, binding) list_with_rest
 
 and object_target_elt =
-  | TargetPropertyId of ident * initialiser option
-  | TargetProperty of property_name * expression
+  | TargetPropertyId of ident_prop * initialiser option
+  | TargetProperty of property_name * expression * initialiser option
   | TargetPropertySpread of expression
   | TargetPropertyMethod of property_name * method_
 
@@ -439,11 +439,11 @@ and assignment_target =
   | ObjectTarget of object_target_elt list
   | ArrayTarget of array_target_elt list
 
-and binding_ident = ident
+and ident_prop = Prop_and_ident of ident
 
 and binding_property =
   | Prop_binding of property_name * binding_element
-  | Prop_ident of binding_ident * initialiser option
+  | Prop_ident of ident_prop * initialiser option
 
 and function_body = statement_list
 
@@ -530,7 +530,7 @@ and bound_idents_of_pattern p =
   match p with
   | ObjectBinding { list; rest } -> (
       List.concat_map list ~f:(function
-          | Prop_ident (i, _) -> [ i ]
+          | Prop_ident (Prop_and_ident i, _) -> [ i ]
           | Prop_binding (_, e) -> bound_idents_of_element e)
       @
       match rest with
@@ -565,8 +565,8 @@ end)
 
 let dot e l = EDot (e, ANormal, l)
 
-let variable_declaration l =
-  Variable_statement (Var, List.map l ~f:(fun (i, e) -> DeclIdent (i, Some e)))
+let variable_declaration ?(kind = Var) l =
+  Variable_statement (kind, List.map l ~f:(fun (i, e) -> DeclIdent (i, Some e)))
 
 let array l = EArr (List.map l ~f:(fun x -> Element x))
 
@@ -588,10 +588,17 @@ let rec assignment_target_of_expr' x =
       let list =
         List.map l ~f:(function
             | Property (PNI n, EVar (S { name = n'; _ } as id))
-              when Utf8_string.equal n n' -> TargetPropertyId (id, None)
-            | Property (n, e) -> TargetProperty (n, assignment_target_of_expr' e)
+              when Utf8_string.equal n n' -> TargetPropertyId (Prop_and_ident id, None)
+            | Property (n, e) ->
+                let e, i =
+                  match e with
+                  | EBin (Eq, e, i) -> e, Some (i, N)
+                  | _ -> e, None
+                in
+                TargetProperty (n, assignment_target_of_expr' e, i)
             | CoverInitializedName (_, i, (e, loc)) ->
-                TargetPropertyId (i, Some (assignment_target_of_expr' e, loc))
+                TargetPropertyId
+                  (Prop_and_ident i, Some (assignment_target_of_expr' e, loc))
             | PropertySpread e -> TargetPropertySpread (assignment_target_of_expr' e)
             | PropertyMethod (n, m) -> TargetPropertyMethod (n, m))
       in
