@@ -65,24 +65,23 @@ let list_files name paths =
     | None -> name, []
   in
   let file =
-    if Filename.is_relative name
-    then
-      try Findlib.find_in_findlib_paths paths name
-      with Not_found -> failwith (Printf.sprintf "file '%s' not found" name)
-    else name
+    match Findlib.find paths name with
+    | None -> failwith (Printf.sprintf "file '%s' not found" name)
+    | Some file -> file
   in
   expand_path exts file virtname
 
 let find_cmi paths base =
-  let name, filename =
-    try
-      let name = String.uncapitalize_ascii base ^ ".cmi" in
-      name, Findlib.find_in_findlib_paths paths name
-    with Not_found ->
-      let name = String.capitalize_ascii base ^ ".cmi" in
-      name, Findlib.find_in_findlib_paths paths name
-  in
-  Filename.concat "/static/cmis" name, filename
+  match
+    List.find_map
+      [ String.uncapitalize_ascii base ^ ".cmi"; String.capitalize_ascii base ^ ".cmi" ]
+      ~f:(fun name ->
+        match Fs.find_in_path paths name with
+        | Some cmi -> Some (name, cmi)
+        | None -> None)
+  with
+  | Some (name, filename) -> Some (Filename.concat "/static/cmis" name, filename)
+  | None -> None
 
 let instr_of_name_content prim ~name ~content =
   let open Code in
@@ -91,23 +90,28 @@ let instr_of_name_content prim ~name ~content =
     | `create_file -> "jsoo_create_file"
     | `create_file_extern -> "jsoo_create_file_extern"
   in
-  Let
-    ( Var.fresh ()
-    , Prim (Extern prim, [ Pc (NativeString name); Pc (NativeString content) ]) )
+  assert (String.is_valid_utf_8 name);
+  ( Let
+      ( Var.fresh ()
+      , Prim
+          ( Extern prim
+          , [ Pc (NativeString (Code.Native_string.of_string name))
+            ; Pc (NativeString (Code.Native_string.of_bytestring content))
+            ] ) )
+  , Code.noloc )
 
 let embed_file ~name ~filename =
   instr_of_name_content `create_file_extern ~name ~content:(Fs.read_file filename)
 
-let init () = Code.(Let (Var.fresh (), Prim (Extern "caml_fs_init", [])))
+let init () = Code.(Let (Var.fresh (), Prim (Extern "caml_fs_init", [])), noloc)
 
 let f ~prim ~cmis ~files ~paths =
   let cmi_files, missing_cmis =
     StringSet.fold
       (fun s (acc, missing) ->
-        try
-          let name, filename = find_cmi paths s in
-          (name, Fs.read_file filename) :: acc, missing
-        with Not_found -> acc, s :: missing)
+        match find_cmi paths s with
+        | Some (name, filename) -> (name, Fs.read_file filename) :: acc, missing
+        | None -> acc, s :: missing)
       cmis
       ([], [])
   in
