@@ -33,7 +33,8 @@ let%expect_test "array_set" =
   print_fun_decl program (Some "some_name");
   [%expect
     {|
-    function some_name(a,n){runtime.caml_check_bound(a,n)[1 + n] = n;return 1} |}]
+    function some_name(a, n){runtime.caml_check_bound(a, n)[1 + n] = n; return 1;}
+    //end |}]
 
 let%expect_test "array_set" =
   compile_and_run array_set;
