@@ -1,6 +1,7 @@
 let () =
   let content = {|
 let () = print_endline "hello";;
+1+1;;
 1+;;
 Missing_module.f;;
 |} in
@@ -12,10 +13,10 @@ Missing_module.f;;
     try
       Location.reset ();
       let phr = !Toploop.parse_toplevel_phrase lexbuf in
-      if not (Toploop.execute_phrase true Format.std_formatter phr) then raise Exit
+      let res = Toploop.execute_phrase true Format.std_formatter phr in
+      flush_all ();
+      if not res then raise Exit
     with
-    | End_of_file ->
-        flush_all ();
-        exit 0
+    | End_of_file -> exit 0
     | x -> Location.report_exception Format.std_formatter x
   done
